@@ -1,10 +1,9 @@
 use anyhow::{Context, Error, Result};
-use bstr::ByteSlice;
 use env_logger::Env;
 use grep_cli::stdout;
 use hcklib::{
     core::{Core, CoreConfig, CoreConfigBuilder, HckInput},
-    field_range::{FieldRange, RegexOrStr},
+    field_range::RegexOrStr,
     line_parser::{RegexLineParser, SubStrLineParser},
     mmap::MmapChoice,
 };
@@ -139,10 +138,20 @@ struct Opts {
     #[structopt(short = "f", long)]
     fields: Option<String>,
 
+    /// Fields to exclude from the output, ex: 3,9-11,15-. Exclude fields are 1 based and inclusive.
+    /// Exclude fields take precedence over `fields`.
+    #[structopt(short = "e", long)]
+    exclude: Option<String>,
+
+    /// Headers to exclude from the output, ex: '^badfield.*$`. This is a string literal by default.
+    /// Add the `-r` flag to treat as a regex.
+    #[structopt(short = "E", long, multiple = true, number_of_values = 1)]
+    exclude_header: Option<Vec<Regex>>,
+
     /// A string literal or regex to select headers, ex: '^is_.*$`. This is a string literal
-    /// by deafult. add the `-r` flag to treat it as a regex.
-    #[structopt(short = "F", long)]
-    header_fields: Option<Vec<Regex>>,
+    /// by default. add the `-r` flag to treat it as a regex.
+    #[structopt(short = "F", long, multiple = true, number_of_values = 1)]
+    header_field: Option<Vec<Regex>>,
 
     /// Treat the header_fields as regexs instead of string literals
     #[structopt(short = "r", long)]
@@ -152,7 +161,7 @@ struct Opts {
     #[structopt(short = "z", long)]
     try_decompress: bool,
 
-    /// Disallow the posibility of using mmap
+    /// Disallow the possibility of using mmap
     #[structopt(long)]
     no_mmap: bool,
 
@@ -190,24 +199,30 @@ fn main() -> Result<()> {
     } else {
         LineTerminator::default()
     };
-    conf_builder.line_terminator(line_term);
+    conf_builder = conf_builder.line_terminator(line_term);
 
     let mmap = if opts.no_mmap {
         MmapChoice::never()
     } else {
         unsafe { MmapChoice::auto() }
     };
-    conf_builder.mmap(mmap);
-    conf_builder.delimiter(&opts.delimiter.as_bytes());
-    conf_builder.output_delimiter(&opts.output_delimiter.as_bytes());
-    conf_builder.is_regex_parser(!opts.delim_is_literal);
-    conf_builder.try_decompress(opts.try_decompress);
-    let conf = conf_builder.build();
+    let conf = conf_builder
+        .mmap(mmap)
+        .delimiter(&opts.delimiter.as_bytes())
+        .output_delimiter(&opts.output_delimiter.as_bytes())
+        .is_regex_parser(!opts.delim_is_literal)
+        .try_decompress(opts.try_decompress)
+        .fields(opts.fields.as_deref())
+        .headers(opts.header_field.as_deref())
+        .exclude(opts.exclude.as_deref())
+        .exclude_headers(opts.exclude_header.as_deref())
+        .header_is_regex(opts.header_is_regex)
+        .build()?;
 
     let mut line_buffer = LineBufferBuilder::new().build();
 
     for input in inputs.into_iter() {
-        if let Err(err) = run(input, &mut writer, &opts, conf, &mut line_buffer) {
+        if let Err(err) = run(input, &mut writer, &conf, &mut line_buffer) {
             if is_broken_pipe(&err) {
                 exit(0)
             }
@@ -222,54 +237,21 @@ fn main() -> Result<()> {
 fn run<W: Write>(
     input: HckInput<PathBuf>,
     writer: &mut W,
-    opts: &Opts,
-    conf: CoreConfig,
+    conf: &CoreConfig,
     line_buffer: &mut LineBuffer,
 ) -> Result<()> {
     let writer = BufWriter::new(writer);
 
-    let delim = if conf.is_parser_regex() {
-        RegexOrStr::Regex(Regex::new(conf.delimiter().to_str()?)?)
-    } else {
-        RegexOrStr::Str(conf.delimiter().to_str()?)
-    };
+    let (extra, fields) = conf.parse_fields(&input)?;
+    // No point processing empty fields
+    if fields.is_empty() {
+        return Ok(());
+    }
 
-    // Parser the fields in the context of the files being looked at
-    let (extra, fields) = match (&opts.fields, &opts.header_fields) {
-        (Some(field_list), Some(header_fields)) => {
-            let first_line = input.peek_first_line()?;
-            let mut fields = FieldRange::from_list(field_list)?;
-            let header_fields = FieldRange::from_header_list(
-                header_fields,
-                first_line.as_bytes(),
-                &delim,
-                opts.header_is_regex,
-            )?;
-            fields.extend(header_fields.into_iter());
-            FieldRange::post_process_ranges(&mut fields);
-            (Some(first_line), fields)
-        }
-        (Some(field_list), None) => (None, FieldRange::from_list(field_list)?),
-        (None, Some(header_fields)) => {
-            let first_line = input.peek_first_line()?;
-            let fields = FieldRange::from_header_list(
-                header_fields,
-                first_line.as_bytes(),
-                &delim,
-                opts.header_is_regex,
-            )?;
-            (Some(first_line), fields)
-        }
-        (None, None) => {
-            eprintln!("Must select one or both `fields` and 'header-fields`.");
-            exit(1);
-        }
-    };
-
-    match &delim {
+    match conf.parsed_delim() {
         RegexOrStr::Regex(regex) => {
             let mut core = Core::new(
-                &conf,
+                conf,
                 &fields,
                 RegexLineParser::new(&fields, &regex),
                 line_buffer,
@@ -314,11 +296,13 @@ mod test {
             delim_is_literal: false,
             output_delimiter: "\t".to_owned(),
             fields: Some(fields.to_owned()),
-            header_fields: None,
+            header_field: None,
             header_is_regex: true,
             try_decompress: false,
             no_mmap,
             crlf: false,
+            exclude: None,
+            exclude_header: None,
         }
     }
 
@@ -337,11 +321,43 @@ mod test {
             delim_is_literal: true,
             output_delimiter: "\t".to_owned(),
             fields: Some(fields.to_owned()),
-            header_fields: None,
+            header_field: None,
             header_is_regex: true,
             try_decompress: false,
             no_mmap,
             crlf: false,
+            exclude: None,
+            exclude_header: None,
+        }
+    }
+
+    /// Build a set of opts for testing
+    #[allow(clippy::too_many_arguments)]
+    fn build_opts_generic(
+        input_file: impl AsRef<Path>,
+        output_file: impl AsRef<Path>,
+        fields: Option<&str>,
+        header_field: Option<Vec<Regex>>,
+        exclude: Option<&str>,
+        no_mmap: bool,
+        delimiter: &str,
+        delim_is_literal: bool,
+        header_is_regex: bool,
+    ) -> Opts {
+        Opts {
+            input: vec![input_file.as_ref().to_path_buf()],
+            output: Some(output_file.as_ref().to_path_buf()),
+            delimiter: delimiter.to_string(),
+            delim_is_literal,
+            output_delimiter: "\t".to_owned(),
+            fields: fields.map(|f| f.to_owned()),
+            header_field,
+            header_is_regex,
+            try_decompress: false,
+            no_mmap,
+            crlf: false,
+            exclude: exclude.map(|e| e.to_owned()),
+            exclude_header: None,
         }
     }
 
@@ -382,14 +398,19 @@ mod test {
                 unsafe { MmapChoice::auto() }
             })
             .output_delimiter(opts.output_delimiter.as_bytes())
-            .build();
+            .headers(opts.header_field.as_deref())
+            .fields(opts.fields.as_deref())
+            .exclude(opts.exclude.as_deref())
+            .exclude_headers(opts.exclude_header.as_deref())
+            .header_is_regex(opts.header_is_regex)
+            .build()
+            .unwrap();
         let mut line_buffer = LineBufferBuilder::new().build();
         let mut writer = BufWriter::new(File::create(output).unwrap());
         run(
             HckInput::Path(input.as_ref().to_owned()),
             &mut writer,
-            opts,
-            conf,
+            &conf,
             &mut line_buffer,
         )
         .unwrap();
@@ -397,6 +418,368 @@ mod test {
 
     const FOURSPACE: &str = "    ";
 
+    #[rstest]
+    fn test_exclude_one(
+        #[values(true, false)] no_mmap: bool,
+        #[values(r" ", "  ")] hck_delim: &str,
+        #[values(true, false)] delim_is_literal: bool,
+    ) {
+        let tmp = TempDir::new().unwrap();
+        let input_file = tmp.path().join("input.txt");
+        let output_file = tmp.path().join("output.txt");
+        let opts = build_opts_generic(
+            &input_file,
+            &output_file,
+            Some("1,3"),
+            None,
+            Some("3"),
+            no_mmap,
+            hck_delim,
+            delim_is_literal,
+            false,
+        );
+        let data = vec![vec!["a", "b", "c"], vec!["1", "2", "3"]];
+        write_file(&input_file, data, hck_delim);
+        run_wrapper(&input_file, &output_file, &opts);
+        let filtered = read_tsv(output_file);
+
+        assert_eq!(filtered, vec![vec!["a",], vec!["1"]]);
+    }
+
+    #[rstest]
+    fn test_exclude_range_overlap_front(
+        #[values(true, false)] no_mmap: bool,
+        #[values(r" ", "  ")] hck_delim: &str,
+        #[values(true, false)] delim_is_literal: bool,
+    ) {
+        let tmp = TempDir::new().unwrap();
+        let input_file = tmp.path().join("input.txt");
+        let output_file = tmp.path().join("output.txt");
+        let opts = build_opts_generic(
+            &input_file,
+            &output_file,
+            Some("3-"),
+            None,
+            Some("-5"),
+            no_mmap,
+            hck_delim,
+            delim_is_literal,
+            false,
+        );
+        let data = vec![
+            vec!["a", "b", "c", "d", "e", "f"],
+            vec!["1", "2", "3", "4", "5", "6"],
+        ];
+        write_file(&input_file, data, hck_delim);
+        run_wrapper(&input_file, &output_file, &opts);
+        let filtered = read_tsv(output_file);
+
+        assert_eq!(filtered, vec![vec!["f",], vec!["6"]]);
+    }
+
+    #[rstest]
+    fn test_exclude_range_overlap_back(
+        #[values(true, false)] no_mmap: bool,
+        #[values(r" ", "  ")] hck_delim: &str,
+        #[values(true, false)] delim_is_literal: bool,
+    ) {
+        let tmp = TempDir::new().unwrap();
+        let input_file = tmp.path().join("input.txt");
+        let output_file = tmp.path().join("output.txt");
+        let opts = build_opts_generic(
+            &input_file,
+            &output_file,
+            Some("2-5"),
+            None,
+            Some("3-"),
+            no_mmap,
+            hck_delim,
+            delim_is_literal,
+            false,
+        );
+        let data = vec![
+            vec!["a", "b", "c", "d", "e", "f"],
+            vec!["1", "2", "3", "4", "5", "6"],
+        ];
+        write_file(&input_file, data, hck_delim);
+        run_wrapper(&input_file, &output_file, &opts);
+        let filtered = read_tsv(output_file);
+
+        assert_eq!(filtered, vec![vec!["b",], vec!["2"]]);
+    }
+
+    #[rstest]
+    fn test_exclude_range_split_fields(
+        #[values(true, false)] no_mmap: bool,
+        #[values(r" ", "  ")] hck_delim: &str,
+        #[values(true, false)] delim_is_literal: bool,
+    ) {
+        let tmp = TempDir::new().unwrap();
+        let input_file = tmp.path().join("input.txt");
+        let output_file = tmp.path().join("output.txt");
+        let opts = build_opts_generic(
+            &input_file,
+            &output_file,
+            Some("1-"),
+            None,
+            Some("3-5"),
+            no_mmap,
+            hck_delim,
+            delim_is_literal,
+            false,
+        );
+        let data = vec![
+            vec!["a", "b", "c", "d", "e", "f"],
+            vec!["1", "2", "3", "4", "5", "6"],
+        ];
+        write_file(&input_file, data, hck_delim);
+        run_wrapper(&input_file, &output_file, &opts);
+        let filtered = read_tsv(output_file);
+
+        assert_eq!(filtered, vec![vec!["a", "b", "f"], vec!["1", "2", "6"]]);
+    }
+
+    #[rstest]
+    fn test_exclude_range_all(
+        #[values(true, false)] no_mmap: bool,
+        #[values(r" ", "  ")] hck_delim: &str,
+        #[values(true, false)] delim_is_literal: bool,
+    ) {
+        let tmp = TempDir::new().unwrap();
+        let input_file = tmp.path().join("input.txt");
+        let output_file = tmp.path().join("output.txt");
+        let opts = build_opts_generic(
+            &input_file,
+            &output_file,
+            Some("4,3"),
+            None,
+            Some("2-5"),
+            no_mmap,
+            hck_delim,
+            delim_is_literal,
+            false,
+        );
+        let data = vec![
+            vec!["a", "b", "c", "d", "e", "f"],
+            vec!["1", "2", "3", "4", "5", "6"],
+        ];
+        write_file(&input_file, data, hck_delim);
+        run_wrapper(&input_file, &output_file, &opts);
+        let filtered = read_tsv(output_file);
+
+        assert!(filtered.is_empty());
+    }
+
+    #[rstest]
+    fn test_exclude_range_split_fields_reorder(
+        #[values(true, false)] no_mmap: bool,
+        #[values(r" ", "  ")] hck_delim: &str,
+        #[values(true, false)] delim_is_literal: bool,
+    ) {
+        let tmp = TempDir::new().unwrap();
+        let input_file = tmp.path().join("input.txt");
+        let output_file = tmp.path().join("output.txt");
+        let opts = build_opts_generic(
+            &input_file,
+            &output_file,
+            Some("4-6,1-3"),
+            None,
+            Some("3-5"),
+            no_mmap,
+            hck_delim,
+            delim_is_literal,
+            false,
+        );
+        let data = vec![
+            vec!["a", "b", "c", "d", "e", "f"],
+            vec!["1", "2", "3", "4", "5", "6"],
+        ];
+        write_file(&input_file, data, hck_delim);
+        run_wrapper(&input_file, &output_file, &opts);
+        let filtered = read_tsv(output_file);
+
+        assert_eq!(filtered, vec![vec!["f", "a", "b"], vec!["6", "1", "2"]]);
+    }
+
+    #[rstest]
+    fn test_headers_simple(
+        #[values(true, false)] no_mmap: bool,
+        #[values(r" ", "  ")] hck_delim: &str,
+        #[values(true, false)] delim_is_literal: bool,
+        #[values(true, false)] header_is_regex: bool,
+    ) {
+        let tmp = TempDir::new().unwrap();
+        let input_file = tmp.path().join("input.txt");
+        let output_file = tmp.path().join("output.txt");
+        let opts = build_opts_generic(
+            &input_file,
+            &output_file,
+            None,
+            Some(vec![Regex::new("a").unwrap()]),
+            None,
+            no_mmap,
+            hck_delim,
+            delim_is_literal,
+            header_is_regex,
+        );
+        let data = vec![
+            vec!["a", "b", "c", "d", "e", "f"],
+            vec!["1", "2", "3", "4", "5", "6"],
+        ];
+        write_file(&input_file, data, hck_delim);
+        run_wrapper(&input_file, &output_file, &opts);
+        let filtered = read_tsv(output_file);
+
+        assert_eq!(filtered, vec![vec!["a"], vec!["1"]]);
+    }
+
+    #[rstest]
+    fn test_headers_simple2(
+        #[values(true, false)] no_mmap: bool,
+        #[values(r" ", "  ")] hck_delim: &str,
+        #[values(true, false)] delim_is_literal: bool,
+        #[values(true, false)] header_is_regex: bool,
+    ) {
+        let tmp = TempDir::new().unwrap();
+        let input_file = tmp.path().join("input.txt");
+        let output_file = tmp.path().join("output.txt");
+        let opts = build_opts_generic(
+            &input_file,
+            &output_file,
+            None,
+            Some(vec![Regex::new("a").unwrap(), Regex::new("c").unwrap()]),
+            None,
+            no_mmap,
+            hck_delim,
+            delim_is_literal,
+            header_is_regex,
+        );
+        let data = vec![vec!["a", "b", "c"], vec!["1", "2", "3"]];
+        write_file(&input_file, data, hck_delim);
+        run_wrapper(&input_file, &output_file, &opts);
+        let filtered = read_tsv(output_file);
+
+        assert_eq!(filtered, vec![vec!["a", "c"], vec!["1", "3"]]);
+    }
+
+    #[rstest]
+    fn test_duplicate_field_selection_more(
+        #[values(true, false)] no_mmap: bool,
+        #[values(r" ", "  ")] hck_delim: &str,
+        #[values(true, false)] delim_is_literal: bool,
+        #[values(true, false)] header_is_regex: bool,
+    ) {
+        let tmp = TempDir::new().unwrap();
+        let input_file = tmp.path().join("input.txt");
+        let output_file = tmp.path().join("output.txt");
+        let opts = build_opts_generic(
+            &input_file,
+            &output_file,
+            Some("3,3,1,2"),
+            None,
+            None,
+            no_mmap,
+            hck_delim,
+            delim_is_literal,
+            header_is_regex,
+        );
+        let data = vec![vec!["a", "b", "c", "d", "e"], vec!["1", "2", "3", "4", "5"]];
+        write_file(&input_file, data, hck_delim);
+        run_wrapper(&input_file, &output_file, &opts);
+        let filtered = read_tsv(output_file);
+
+        assert_eq!(filtered, vec![vec!["c", "a", "b"], vec!["3", "1", "2"]]);
+    }
+
+    #[rstest]
+    fn test_duplicate_field_selection_range(
+        #[values(true, false)] no_mmap: bool,
+        #[values(r" ", "  ")] hck_delim: &str,
+        #[values(true, false)] delim_is_literal: bool,
+        #[values(true, false)] header_is_regex: bool,
+    ) {
+        let tmp = TempDir::new().unwrap();
+        let input_file = tmp.path().join("input.txt");
+        let output_file = tmp.path().join("output.txt");
+        let opts = build_opts_generic(
+            &input_file,
+            &output_file,
+            Some("2-3,5,1,2-4"),
+            None,
+            None,
+            no_mmap,
+            hck_delim,
+            delim_is_literal,
+            header_is_regex,
+        );
+        let data = vec![vec!["a", "b", "c", "d", "e"], vec!["1", "2", "3", "4", "5"]];
+        write_file(&input_file, data, hck_delim);
+        run_wrapper(&input_file, &output_file, &opts);
+        let filtered = read_tsv(output_file);
+
+        assert_eq!(
+            filtered,
+            vec![vec!["b", "c", "e", "a", "d"], vec!["2", "3", "5", "1", "4"]]
+        );
+    }
+
+    #[rstest]
+    fn test_headers_and_fields(
+        #[values(true, false)] no_mmap: bool,
+        #[values(r" ", "  ")] hck_delim: &str,
+        #[values(true, false)] delim_is_literal: bool,
+        #[values(true, false)] header_is_regex: bool,
+    ) {
+        let tmp = TempDir::new().unwrap();
+        let input_file = tmp.path().join("input.txt");
+        let output_file = tmp.path().join("output.txt");
+        let opts = build_opts_generic(
+            &input_file,
+            &output_file,
+            Some("3"),
+            Some(vec![Regex::new("b").unwrap(), Regex::new("a").unwrap()]),
+            None,
+            no_mmap,
+            hck_delim,
+            delim_is_literal,
+            header_is_regex,
+        );
+        let data = vec![vec!["a", "b", "c", "d", "e"], vec!["1", "2", "3", "4", "5"]];
+        write_file(&input_file, data, hck_delim);
+        run_wrapper(&input_file, &output_file, &opts);
+        let filtered = read_tsv(output_file);
+
+        assert_eq!(filtered, vec![vec!["b", "c", "a"], vec!["2", "3", "1"]]);
+    }
+
+    #[rstest]
+    fn test_duplicate_field_selection(
+        #[values(true, false)] no_mmap: bool,
+        #[values(r" ", "  ")] hck_delim: &str,
+        #[values(true, false)] delim_is_literal: bool,
+        #[values(true, false)] header_is_regex: bool,
+    ) {
+        let tmp = TempDir::new().unwrap();
+        let input_file = tmp.path().join("input.txt");
+        let output_file = tmp.path().join("output.txt");
+        let opts = build_opts_generic(
+            &input_file,
+            &output_file,
+            Some("3,1,3"),
+            None,
+            None,
+            no_mmap,
+            hck_delim,
+            delim_is_literal,
+            header_is_regex,
+        );
+        let data = vec![vec!["a", "b", "c", "d"], vec!["1", "2", "3", "4"]];
+        write_file(&input_file, data, hck_delim);
+        run_wrapper(&input_file, &output_file, &opts);
+        let filtered = read_tsv(output_file);
+
+        assert_eq!(filtered, vec![vec!["c", "a"], vec!["3", "1"]]);
+    }
     #[rstest]
     #[rustfmt::skip::macros(vec)]
     fn test_read_single_values(
